"""Tests ``execution`` module."""

import pytest

from uranie_launcher import execution



## execution
def test_execution_default():
    """test exec default"""

    exe = execution.Execution(working_directory=".")

    assert (
        exe.working_directory == "." and
        exe.clean is True and
        exe.save == -1 and
        exe.visualization is False
    )


def test_execution_enable_visualization():
    """test exec enable visualization"""

    exe = execution.Execution(working_directory=".")
    exe.enable_visualization()
    exe.clean_outputs(False)
    exe.save_outputs(0)

    assert (
        exe.clean is False and
        exe.save == 0 and
        exe.visualization is True
    )


def test_execution_local():
    """test local exec"""

    with pytest.raises(ValueError) as error:
        execution.ExecutionLocal(working_directory=".", nb_jobs=0)
    assert "nb_jobs' (0) must be >=1." in str(error.value)

    nb_jobs = 2
    exe = execution.ExecutionLocal(working_directory="test_execution_local", nb_jobs=nb_jobs)

    assert (
        exe.visualization is False and
        exe.nb_jobs == nb_jobs
    )


def test_execution_slurm_raise():
    """test slurm exec"""

    with pytest.raises(NotImplementedError) as error:
        execution.ExecutionSlurm()
    assert "Execution on cluster is not implemented yet" in str(error.value)



"""Class used to set execution info into an understandable format for uranie_launcher.
"""
from pathlib import Path


class Execution:
    """ Parent class of classes describing how to execute the calculations.
    """
    def __init__(self, working_directory: Path) -> None:
        """Constructor.

        Parameters
        ----------
        working_directory : Path
            Path to the working directory of Uranie Launcher.
        """
        self._working_directory = working_directory
        self._visualization = False
        self._clean = True
        self._save = -1

    @property
    def working_directory(self) -> Path:
        """Path to the working directory of Uranie Launcher.

        Returns
        -------
        Path
            Path to the directory
        """
        return self._working_directory

    @property
    def visualization(self) -> bool:
        """Get the visialisation status

        Returns
        -------
        bool
            Enabled if True
        """
        return self._visualization

    @property
    def clean(self) -> bool:
        """Get the cleanning status

        Returns
        -------
        bool
            Enabled if True
        """
        return self._clean

    @property
    def save(self) -> int:
        """Get the saving status

        Returns
        -------
        bool
            Enabled if True
        """
        return self._save

    def enable_visualization(self, enable: bool = True):
        """To enable/disable visualization

        Parameters
        ----------
        enable : bool, optional
            True to enable, by default True
        """
        self._visualization = enable

    def clean_outputs(self, enable: bool = True):
        """To enable/disable cleanning

        From Uranie documentation: set the clean flag to the value true or false. If it is set to
        True, all the working directories are cleaned before launching. Then if the save flag is
        set to True, the working directories are cleaned a second time after the TLauncher run.
        If not specified, default parameter is kTrue, which means the working directory will be
        cleaned (i.e. temporary folders will be erased)

        Parameters
        ----------
        enable : bool, optional
            True to enable, by default True
        """
        self._clean = enable

    def save_outputs(self, enable: int = -1):
        """To enable/disable saving

        From Uranie documentation: activate the save mode and set the maximum number i of
        folders to be kept. A positive value for i stands for the explicit maximum number of saves,
        while no value or -1 value allow to keep all the patterns (all patterns are launched in a
        different working directory UranieLauncher_i). Default parameter is -1.
        If setSave is not called, the temporary output folders of the code will be overwritten at
        each computing step, while they will be left unmodified if the method is called.

        Example
        -------
        >>> mylauncher.setSave() # will save each launch
        >>> mylauncher.setSave(-1) # will save each launch
        >>> mylauncher.setSave(10) # will save 10 launches

        Warning
        -------
        In case of multi-process distributed computation, in order to prevent all the processes
        from trying and accessing the same data at the same time, a call to the setSave method is
        mandatory. As a consequence, if it is not manually set, Uranie will set it automatically

        Parameters
        ----------
        enable : int, optional
            True to enable, by default -1
        """
        self._save = enable


class ExecutionLocal(Execution):
    """Object containing the info about how to execute the calculations on a desktop.
    """
    def __init__(self, working_directory: Path, nb_jobs: int = 1) -> None:
        """Constructor.

        Parameters
        ----------
        working_directory : Path
            Path to the working directory of Uranie Launcher.
        nb_jobs : int, optional
            Number of parallel executions, by default 1.

        Raises
        ------
        ValueError
            'nb_jobs' must be >=1.
        """
        super().__init__(working_directory=working_directory)
        if nb_jobs < 1:
            raise ValueError(f"'nb_jobs' ({nb_jobs}) must be >=1.")
        self._nb_jobs = nb_jobs

    @property
    def nb_jobs(self):
        """Get the number of jobs

        Returns
        -------
        int
            value >0
        """
        return self._nb_jobs


class ExecutionSlurm(Execution):
    """Object containing the info about how to execute the calculations on a cluster using SLURM.
    """
    def __init__(self):
        super().__init__(working_directory="")
        # TODO implementer l'execution sur cluster
        raise NotImplementedError("Execution on cluster is not implemented yet")
#         self._nb_of_tasks = 1
#         self._nb_of_cpu_per_task = 10
#         self._memory_per_cpu = "100M"
#         self._partitions = "amdq_naples,amdq_rome,intelq_5118,intelq_6226,intelq_6234"
#         self._qos = "normal"
#         self._account = ""
#         self._walltime = "01:00:00"
#         self._job_name = "my_job"
#         self._output_file = "output_file.out"
#         self._error_file = "error_file"
#         self._email = "your.name@example.com"
#         self._email_type = "ALL"
#         self._liste_of_nodes = ""
#         self._myscript_to_launch = "myscript.sh"

#     def set_nb_of_tasks(self, nb_of_tasks: int):
#         """Set the number of tasks

#         Parameters
#         ----------
#         execution_mode : int
#             Set _nb_of_tasks
#         """
#         self._nb_of_tasks = nb_of_tasks

#     @property
#     def nb_of_tasks(self):
#         """Get the number of tasks

#         Returns
#         -------
#         int
#             self._nb_of_tasks
#         """
#         return self._nb_of_tasks


#     def set_nb_of_cpu_per_task(self, nb_of_cpu_per_task: int):
#         """Set the number of CPU per tasks

#         Parameters
#         ----------
#         execution_mode : int
#             Set _nb_of_cpu_per_task
#         """
#         self._nb_of_cpu_per_task = nb_of_cpu_per_task

#     @property
#     def nb_of_cpu_per_task(self):
#         """Get the number of CPU per tasks

#         Returns
#         -------
#         int
#             self._nb_of_cpu_per_task
#         """
#         return self._nb_of_cpu_per_task


#     def set_memory_per_cpu(self, memory_per_cpu: str):
#         """Set the memory per cpu

#         Parameters
#         ----------
#         execution_mode : str
#             Set _memory_per_cpu
#         """
#         self._memory_per_cpu = memory_per_cpu

#     @property
#     def memory_per_cpu(self):
#         """Get the memory per cpu

#         Returns
#         -------
#         str
#             self._memory_per_cpu
#         """
#         return self._memory_per_cpu


#     def set_partitions(self, partitions: str):
#         """Set the partitions

#         Parameters
#         ----------
#         execution_mode : str
#             Set _partitions
#         """
#         self._partitions = partitions

#     @property
#     def partitions(self):
#         """Get the partitions

#         Returns
#         -------
#         str
#             self._partitions
#         """
#         return self._partitions


#     def set_qos(self, qos: str):
#         """Set the QoS (Quality of Service)

#         Parameters
#         ----------
#         execution_mode : str
#             Set _qos
#         """
#         self._qos = qos

#     @property
#     def qos(self):
#         """Get the QoS (Quality of Service)

#         Returns
#         -------
#         str
#             self._qos
#         """
#         return self._qos


#     def set_account(self, account: str):
#         """Set the account

#         Parameters
#         ----------
#         execution_mode : str
#             Set _account
#         """
#         self._account = account

#     @property
#     def account(self):
#         """Get the account

#         Returns
#         -------
#         str
#             self._account
#         """
#         return self._account

#     def set_walltime(self, walltime: str):
#         """Set the walltime (format : HH:MM:SS)

#         Parameters
#         ----------
#         execution_mode : str
#             Set _walltime
#         """
#         self._walltime = walltime

#     @property
#     def walltime(self):
#         """Get the walltime

#         Returns
#         -------
#         str
#             self._walltime
#         """
#         return self._walltime


#     def set_job_name(self, job_name: str):
#         """Set the job name

#         Parameters
#         ----------
#         execution_mode : str
#             Set _job_name
#         """
#         self._job_name = job_name

#     @property
#     def job_name(self):
#         """Get the job name

#         Returns
#         -------
#         str
#             self._job_name
#         """
#         return self._job_name


#     def set_output_file(self, output_file: str):
#         """Set the output file name

#         Parameters
#         ----------
#         execution_mode : str
#             Set _output_file
#         """
#         self._output_file = output_file

#     @property
#     def output_file(self):
#         """Get the output file name

#         Returns
#         -------
#         str
#             self._output_file
#         """
#         return self._output_file


#     def set_error_file(self, error_file: str):
#         """Set the error file name

#         Parameters
#         ----------
#         execution_mode : str
#             Set _error_file
#         """
#         self._error_file = error_file

#     @property
#     def error_file(self):
#         """Get the error file name

#         Returns
#         -------
#         str
#             self._error_file
#         """
#         return self._error_file


#     def set_email(self, email: str):
#         """Set the email by which you will be notified

#         Parameters
#         ----------
#         execution_mode : str
#             Set _email
#         """
#         self._email = email

#     @property
#     def email(self):
#         """Get the email by which you will be notified

#         Returns
#         -------
#         str
#             self._email
#         """
#         return self._email


#     def set_email_type(self, email_type: str):
#         """Set the type of event for which you want to be notified by email

#         Parameters
#         ----------
#         execution_mode : str
#             Set _email_type
#         """
#         self._email_type = email_type

#     @property
#     def email_type(self):
#         """Get the type of event for which you want to be notified by email

#         Returns
#         -------
#         str
#             self._email_type
#         """
#         return self._email_type


#     def set_liste_of_nodes(self, liste_of_nodes: str):
#         """_summary_

#         Parameters
#         ----------
#         execution_mode : str
#             Set _liste_of_nodes
#         """
#         self._liste_of_nodes = liste_of_nodes

#     @property
#     def liste_of_nodes(self):
#         """_summary_

#         Returns
#         -------
#         str
#             self._liste_of_nodes
#         """
#         return self._liste_of_nodes


#     def set_myscript_to_launch(self, myscript_to_launch: str):
#         """Set the name of the script to launch

#         Parameters
#         ----------
#         execution_mode : str
#             Set _myscript_to_launch
#         """
#         self._myscript_to_launch = myscript_to_launch

#     @property
#     def myscript_to_launch(self):
#         """Get the name of the script to launch

#         Returns
#         -------
#         str
#             self._myscript_to_launch
#         """
#         return self._myscript_to_launch



""" Script to launch Uranie from the bench.
"""

import json
import logging
import subprocess
import sys
import argparse
from pathlib import Path
from uranie_launcher import __name__ as uranie_launcher_name
from uranie_launcher import utils


def main_unitary(arguments):
    """ Run the unitary function with the given command line arguments.

    Parameters
    ----------
    arguments: list of str
        Command line arguments.

    Returns
    -------
    int
        Return code.
    """
    # Interpret arguments
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "commands_json_file",
        help="Name of the script or the command to execute")
    parser.add_argument(
        "--debug",
        action="store_true",
        help="activate debug mode (default: %(default)s)")
    args = parser.parse_args(arguments)

    utils.set_verbosity(utils.DEBUG if args.debug else utils.INFO)

    with open(args.commands_json_file, encoding='utf-8') as f:
        commands = json.load(f)

    for _command, _arguments in commands:
        utils.info(f"{Path(__file__).name} : Start the execution of the command "
                   f"subprocess.run({[_command] + _arguments})")
        proc = subprocess.run([_command] + _arguments,  # pylint: disable=subprocess-run-check
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)

        utils.info(f"{_command} : proc.stdout:")
        utils.info(proc.stdout.decode(encoding='utf-8'))
        utils.info(f"{_command} : proc.stderr:")
        utils.info(proc.stderr.decode(encoding='utf-8'))
        if proc.returncode == 0:
            utils.info(f"The execution of the command "
                       f"subprocess.run({[_command] + _arguments}) was successful.")
        else:
            utils.info(f"Failed to execute the command "
                       f"subprocess.run({[_command] + _arguments}). Returned {proc.returncode}.")
            return proc.returncode

    return 0


def run_unitary():
    """ Entry point for the ``uranie-launcher-unitary`` script."""
    logging.getLogger(uranie_launcher_name).addHandler(logging.StreamHandler(sys.stdout))
    sys.exit(main_unitary(sys.argv[1:]))



"""Tests ``utils`` module."""

import pytest

from uranie_launcher import utils


def test_set_verbosity():
    """Test log level"""

    new_log_level = 2
    utils.set_verbosity(new_log_level)

    assert (
        utils.get_log_level() == new_log_level
    )


def test_set_verbosity_lower_that_min_bound():
    """Test log level"""

    new_log_level = -1
    old_log_level = utils.get_log_level()

    with pytest.raises(ValueError) as error:
        utils.set_verbosity(new_log_level)
    assert (
        f"log_level must be in [{utils.NONE};{utils.DEBUG}]" in str(error.value)
    )

    assert old_log_level == utils.get_log_level()


def test_set_verbosity_greater_that_upper_bound():
    """Test log level"""

    new_log_level = 3
    old_log_level = utils.get_log_level()

    with pytest.raises(ValueError) as error:
        utils.set_verbosity(new_log_level)
    assert (
        f"log_level must be in [{utils.NONE};{utils.DEBUG}]" in str(error.value)
    )

    assert old_log_level == utils.get_log_level()


def test_log(capsys):
    """Test log level"""

    new_log_level = 1
    utils.set_verbosity(new_log_level)
    message_log = "un message de log"
    utils.log(utils.INFO, message_log)

    # Capture output
    captured_log = capsys.readouterr()

    assert (
        utils.get_log_level() == new_log_level and
        captured_log.out == f"{message_log}\n"
    )


def test_no_info_output_when_log_level_below_info(capsys):
    """Test log outputs"""

    new_log_level = utils.NONE
    utils.set_verbosity(new_log_level)

    message_info = "un message d'info"
    utils.info(message_info)

    # Capture output
    captured_info = capsys.readouterr()

    message_debug = "un message de debug"
    utils.debug(message_debug)

    # Capture output
    captured_debug = capsys.readouterr()

    assert (
        utils.get_log_level() == new_log_level and
        captured_info.out == "" and
        captured_debug.out == ""
        )


def test_no_debug_output_when_log_level_below_debug(capsys):
    """Test log outputs"""

    new_log_level = utils.INFO
    utils.set_verbosity(new_log_level)

    message_info = "un message d'info"
    utils.info(message_info)

    # Capture output
    captured_info = capsys.readouterr()

    message_debug = "un message de debug"
    utils.debug(message_debug)

    # Capture output
    captured_debug = capsys.readouterr()

    assert (
        utils.get_log_level() == new_log_level and
        captured_info.out == f"{message_info}\n" and
        captured_debug.out == ""
        )


def test_all_output_when_log_level_set_to_debug(capsys):
    """Test log outputs"""

    new_log_level = utils.DEBUG
    utils.set_verbosity(new_log_level)

    message_info = "un message d'info"
    utils.info(message_info)

    # Capture output
    captured_info = capsys.readouterr()

    message_debug = "un message de debug"
    utils.debug(message_debug)

    # Capture output
    captured_debug = capsys.readouterr()

    assert (
        utils.get_log_level() == new_log_level and
        captured_info.out == f"{message_info}\n" and
        captured_debug.out == f"{message_debug}\n"
        )



"""Tests ``data`` module."""

from pathlib import Path

import pytest

from uranie_launcher import data


def test_conversion():
    """Test type conversion"""

    assert 1.0 == data.Data.Types.convert("1.0", data.Data.Types.DOUBLE)
    assert "azerty" == data.Data.Types.convert("azerty", data.Data.Types.STRING)
    assert [1.0, 2.0 ,3.0] == data.Data.Types.convert("[1, 2 ,3]", data.Data.Types.VECTOR)


def test_header():
    """Test data header content"""

    header = data.Data.Header(
        name="test_header", value_type=data.Data.Types.DOUBLE, value_unit="None")

    assert header.name == "test_header"
    assert header.value_type == data.Data.Types.DOUBLE
    assert header.value_unit == "None"


def test_data(simple_data: data.Data):
    """Test data content"""

    with pytest.raises(ValueError) as error:
        simple_data.add_values(values=["toto", 1.0, [1.0, 2.0]])
    assert "Type is not correct: found" in str(error.value)

    with pytest.raises(ValueError) as error:
        simple_data.add_values(values=[1.0, "toto", [1.0, "2.0"]])
    assert "Element of vector is not correct: found" in str(error.value)

    assert simple_data.name == "test_data"
    assert simple_data.description == "test"
    assert simple_data.headers
    assert simple_data.names == ["x", "y", "z"]
    assert simple_data.types == [
        data.Data.Types.DOUBLE, data.Data.Types.STRING, data.Data.Types.VECTOR]
    assert simple_data.units == ["u_x", "u_y", "u_z"]
    assert simple_data.values == [[1.0], ["toto"], [[1.0, 2.0]]]
    assert simple_data.nb_rows == 1
    assert simple_data.nb_columns == 3
    assert simple_data.get_values(index=0) == [1.0, "toto", [1.0, 2.0]]


def test_data_to_csv(simple_data: data.Data):
    """Test conversion data <-> csv"""

    output_dirname = Path(__file__).absolute().parent / "test_data_to_csv"
    output_dirname.mkdir(parents=True, exist_ok=True)

    csv_filepath = output_dirname / "simple_data.csv"
    data.data_to_csv(data=simple_data, filepath=csv_filepath)

    simple_data_2 = data.csv_to_data(csv_filepath)

    assert simple_data.name == simple_data_2.name
    assert simple_data.description == simple_data_2.description
    assert len(simple_data.headers) == len(simple_data_2.headers)
    assert simple_data.names == simple_data_2.names
    assert simple_data.types == simple_data_2.types
    assert simple_data.units == simple_data_2.units
    assert simple_data.values == simple_data_2.values
    assert simple_data.nb_rows == simple_data_2.nb_rows
    assert simple_data.nb_columns == simple_data_2.nb_columns
    assert simple_data.get_values(index=0) == simple_data_2.get_values(index=0)


def test_data_to_json(simple_data: data.Data):
    """Test conversion data <-> json"""

    output_dirname = Path(__file__).absolute().parent / "test_data_to_json"
    output_dirname.mkdir(parents=True, exist_ok=True)

    json_filepath = output_dirname / "simple_data.json"
    data.data_to_json(data=simple_data, filepath=json_filepath)

    simple_data_2 = data.json_to_data(json_filepath)

    assert simple_data.name == simple_data_2.name
    assert simple_data.description == simple_data_2.description
    assert len(simple_data.headers) == len(simple_data_2.headers)
    assert simple_data.names == simple_data_2.names
    assert simple_data.types == simple_data_2.types
    assert simple_data.units == simple_data_2.units
    assert simple_data.values == simple_data_2.values
    assert simple_data.nb_rows == simple_data_2.nb_rows
    assert simple_data.nb_columns == simple_data_2.nb_columns
    assert simple_data.get_values(index=0) == simple_data_2.get_values(index=0)


def test_data_to_ascii(simple_data: data.Data):
    """Test conversion data <-> ascii"""

    output_dirname = Path(__file__).absolute().parent / "test_data_to_ascii"
    output_dirname.mkdir(parents=True, exist_ok=True)

    ascii_filepath = output_dirname / "simple_data.dat"
    data.data_to_ascii(data=simple_data, filepath=ascii_filepath)

    simple_data_2 = data.ascii_to_data(ascii_filepath)

    assert simple_data.name == simple_data_2.name
    assert simple_data.description == simple_data_2.description
    assert len(simple_data.headers) == len(simple_data_2.headers)
    assert simple_data.names == simple_data_2.names
    assert simple_data.types == simple_data_2.types
    assert simple_data.units == simple_data_2.units
    assert simple_data.values == simple_data_2.values
    assert simple_data.nb_rows == simple_data_2.nb_rows
    assert simple_data.nb_columns == simple_data_2.nb_columns
    assert simple_data.get_values(index=0) == simple_data_2.get_values(index=0)


def test_no_meta_data():
    """Test file without meta data"""

    output_dirname = Path(__file__).absolute().parent / "test_no_meta_data"
    output_dirname.mkdir(parents=True, exist_ok=True)

    ascii_filepath = output_dirname / "no_meta_data.dat"
    ascii_filepath.write_text("1.0 2.0 3.0", encoding="utf-8")

    no_meta_data = data.ascii_to_data(ascii_filepath)

    ascii_filepath_2 = output_dirname / "no_meta_data_2.dat"
    data.data_to_ascii(data=no_meta_data, filepath=ascii_filepath_2)



"""Tests ``_data_2_uranie`` module."""

from pathlib import Path
import pytest

from URANIE import DataServer, Sampler

from uranie_launcher import execution, data, input_data, _data_2_uranie

## create_data_server
def test_create_data_server(t_data_server, data_server_name, quantity_of_interest):
    """Test create t_data_server"""

    assert (
        isinstance(t_data_server, DataServer.TDataServer) and
        str(t_data_server) == (f"Name: {data_server_name} "
                               f"Title: Quantities = ['{quantity_of_interest}']") and
        t_data_server.GetName() == data_server_name and
        t_data_server.GetTitle() == f"Quantities = ['{quantity_of_interest}']"
    )


## set_inputs
def test_set_inputs_distribution_unknown(t_data_server, simulation_variable, tag_filename):
    """Test distribution raise"""

    inputs = input_data.Inputs()
    inputs.add_input(input_data.Inputs.Input(variable_name=simulation_variable, distribution=None))
    inputs.set_file_flag(Path(__file__).absolute().parent / tag_filename)

    with pytest.raises(ValueError) as error:
        _data_2_uranie.set_inputs(inputs, t_data_server)
    assert "Invalid distribution type: NoneType" in str(error.value)


def test_set_inputs_distribution_uniform(
        t_data_server, data_input_inputs_distribution_uniform, simulation_variable):
    """Test distribution"""


    _data_2_uranie.set_inputs(data_input_inputs_distribution_uniform, t_data_server)

    assert (
        isinstance(t_data_server.getAttribute(simulation_variable),
                   DataServer.TUniformDistribution) and
        str(t_data_server.getAttribute(simulation_variable)) == (
            f"Name: {simulation_variable} Title: {simulation_variable}") and
        t_data_server.getAttribute(simulation_variable).GetName() == simulation_variable and
        t_data_server.getAttribute(simulation_variable).GetTitle() == simulation_variable and
        t_data_server.getAttribute(simulation_variable).getLowerBound() == 100 and
        t_data_server.getAttribute(simulation_variable).getUpperBound() == 1000
        # test the filFlag value...
    )


def test_set_inputs_distribution_truncated_normal(t_data_server,
                                                  data_input_inputs_distribution_truncated_normal,
                                                  simulation_variable):
    """Test distribution"""

    _data_2_uranie.set_inputs(data_input_inputs_distribution_truncated_normal, t_data_server)

    assert (
        isinstance(t_data_server.getAttribute(simulation_variable),
                   DataServer.TNormalDistribution) and
        str(t_data_server.getAttribute(simulation_variable)) == (
            f"Name: {simulation_variable} Title: {simulation_variable}") and
        t_data_server.getAttribute(simulation_variable).GetName() == simulation_variable and
        t_data_server.getAttribute(simulation_variable).GetTitle() == simulation_variable and
        t_data_server.getAttribute(simulation_variable).getLowerBound() == 100 and
        t_data_server.getAttribute(simulation_variable).getUpperBound() == 1000 and
        t_data_server.getAttribute(simulation_variable).getParameterMu() == 600 and
        t_data_server.getAttribute(simulation_variable).getParameterSigma() == 100
        # test the filFlag value...
    )

## generate_sample
def test_generate_sobol_sample(generate_sample, t_data_server, data_input_outputs):
    """Test sobol sample"""

    output_dirname = Path(__file__).absolute().parent / "test_generate_sobol_sample"
    output_dirname.mkdir(parents=True, exist_ok=True)

    sampler = generate_sample(output_dirname,
                              t_data_server,
                              input_data.Propagation.SOBOL,
                              4)

    assert (
        isinstance(sampler, Sampler.TQMC) and
        sampler.getMethodName() == "qMC_sobol" and
        sampler.GetName() == "Sampling_qMC_sobol_4" and
        sampler.GetTitle() == "Uranie 4 sample with qMC method sobol"
    )

    assert ("execution_index" in
            data.ascii_to_data(
                output_dirname / data_input_outputs.experimental_design_filename).names)


def test_generate_srs_sample(generate_sample, t_data_server):
    """Test srs sample"""

    sampler = generate_sample(Path("some/path/to/output_dirname"),
                              t_data_server,
                              input_data.Propagation.SRS,
                              4)

    assert (
        isinstance(sampler, Sampler.TSampling) and
        sampler.getMethodName() == "SRS" and
        sampler.GetName() == "Sampling_srs_4" and
        sampler.GetTitle() == "Uranie 4 sample with method srs"
    )


def test_generate_unknown_sample(generate_sample, t_data_server):
    """Test sample raise"""

    with pytest.raises(ValueError) as error:
        generate_sample(Path("some/path/to/output_dirname"),
                        t_data_server,
                        "unknown_method",
                        4)
    assert "Unknown sampling method: unknown_method" in str(error.value)

@pytest.mark.skip(reason="Not implemented yet")
def test_visualisation(t_data_server, mocker):
    """Test visualisation"""
    # pylint: disable=no-member,import-outside-toplevel

    import ROOT

    # Mock les méthodes et les objets nécessaires
    mocker.patch.object(ROOT.gROOT, 'SetBatch')
    mocker.patch.object(ROOT, 'TCanvas')
    mocker.patch.object(t_data_server, 'drawPairs')

    # Appeler la fonction à tester
    _data_2_uranie.visualisation(t_data_server)

    # Vérifier l'appel des méthodes et des objets
    ROOT.gROOT.SetBatch.assert_called_once_with(False)
    ROOT.TCanvas.assert_called_once_with()
    t_data_server.drawPairs.assert_called_once_with()


def test_set_outputs(t_output_files, result_filename):
    """Test set_outputs"""

    assert (str(t_output_files[0]) ==
            f"Name: {result_filename} Title: TCodeFile with name[{result_filename}]")


def test_create_launcher(generate_t_launcher):
    """Test create t_launcher"""

    output_dirname = Path(__file__).absolute().parent / "test_create_launcher"
    output_dirname.mkdir(parents=True, exist_ok=True)

    generate_t_launcher(output_dirname)


def test_run_calculation_local_parallel(generate_t_launcher):
    """Test run parallel"""

    output_dirname = Path(__file__).absolute().parent / "test_run_calculation_local_parallel"
    output_dirname.mkdir(parents=True, exist_ok=True)

    t_launcher = generate_t_launcher(output_dirname)

    exe = execution.ExecutionLocal(working_directory=output_dirname / "unitary", nb_jobs=2)

    _data_2_uranie.run_calculations(execution=exe,
                                    t_launcher=t_launcher,
                                    output_directory=output_dirname)

    assert (
        t_launcher.getWorkingDirectory() == str(output_dirname / "unitary") and
        t_launcher.getSave() is True and
        t_launcher.getClean() is True
    )


def test_run_calculation_local_sequentiel(generate_t_launcher):
    """Test run suquential"""

    output_dirname = Path(__file__).absolute().parent / "test_run_calculation_local_sequentiel"
    output_dirname.mkdir(parents=True, exist_ok=True)

    t_launcher = generate_t_launcher(output_dirname)

    exe = execution.ExecutionLocal(working_directory=output_dirname / "unitary", nb_jobs=1)

    _data_2_uranie.run_calculations(execution=exe,
                                    t_launcher=t_launcher,
                                    output_directory=output_dirname)

    assert (
        t_launcher.getWorkingDirectory() == str(output_dirname / "unitary") and
        t_launcher.getSave() is True and
        t_launcher.getClean() is True
    )


def test_run_calculation_raise(generate_t_launcher):
    """Test run with raise"""

    output_dirname = Path(__file__).absolute().parent / "test_run_calculation_raise"
    output_dirname.mkdir(parents=True, exist_ok=True)

    t_launcher = generate_t_launcher(output_dirname)

    with pytest.raises(ValueError) as error:
        exe = execution.ExecutionLocal(working_directory=output_dirname, nb_jobs=1)
        _data_2_uranie.run_calculations(execution=exe,
                                        t_launcher=t_launcher,
                                        output_directory=output_dirname)

    assert ("execution.clean is True and output_directory =="
            " execution.working_directory is not possible.") in str(error.value)

    with pytest.raises(ValueError) as error:
        exe = execution.Execution(working_directory=output_dirname / "unitary")
        _data_2_uranie.run_calculations(execution=exe,
                                        t_launcher=t_launcher,
                                        output_directory=output_dirname)

    assert "Invalid execution mode: Execution" in str(error.value)


def test_save_calculations(generate_final_results):
    """Test save without fail"""

    output_dirname = Path(__file__).absolute().parent / "test_save_calculations"
    output_dirname.mkdir(parents=True, exist_ok=True)

    output_file, nb_fails = generate_final_results(output_dirname=output_dirname, sample_size=4)

    assert output_file.is_file()
    assert nb_fails == 0


def test_save_calculations_1_fail(generate_final_results):
    """Test save with fail"""

    output_dirname = Path(__file__).absolute().parent / "test_save_calculations_1_fail"
    output_dirname.mkdir(parents=True, exist_ok=True)
    output_file, nb_fails = generate_final_results(output_dirname=output_dirname, sample_size=10)

    assert output_file.is_file()
    assert nb_fails == 1



"""Define the different verbosity of the logs.
"""

_log_level = 1  # pylint: disable=invalid-name
"""Verbosity level"""

NONE = 0
INFO = 1
DEBUG = 2


def set_verbosity(log_level: int):
    """Define verbosity level.

    Parameters
    ----------
    log_level : int
        log level

    Raises
    ------
    ValueError
        if not in range [NONE;DEBUG]
    """
    global _log_level  # pylint: disable=invalid-name,global-statement
    if not NONE <= log_level <= DEBUG:
        raise ValueError(f"log_level must be in [{NONE};{DEBUG}]")
    _log_level = log_level


def get_log_level():
    """Access to current log level."""
    return _log_level


def log(level, *args, **kwargs):
    """log basis function"""
    if level >= DEBUG and "flush" not in kwargs:
        kwargs["flush"] = True
    if _log_level >= level:
        print(*args, **kwargs)


def info(*args, **kwargs):
    """Log info level. Use it as print function."""
    log(INFO, *args, **kwargs)


def debug(*args, **kwargs):
    """Log debug level. Use it as print function."""
    log(DEBUG, *args, **kwargs)



# -*- coding: utf-8 -*-

"""
This module allow you to use URANIE to run uncertainty propagation
with your own calculation scripts.
"""
import os

import ROOT

__author__ = "Clement STUTZ"
__contact__ = "clement.stutz@cea.fr"
__copyright__ = "Copyright 2023, CEA"
__status__ = "Prototype"

with open(os.path.join(os.path.dirname(__file__), "VERSION"), encoding='utf-8') as file:
    __version__ = file.read().strip()


def _rootlogon():
    """Setup done in rootlogon.py recommanded by """
    # pylint: disable=no-member,invalid-name
    # General graphical style
    WHITE = 0

    # PlotStyle
    ROOT.gStyle.SetPalette(1)
    ROOT.gStyle.SetOptDate(21)

    # Legend
    ROOT.gStyle.SetLegendBorderSize(0)
    ROOT.gStyle.SetFillStyle(0)

    # Pads
    ROOT.gStyle.SetPadColor(WHITE)
    ROOT.gStyle.SetTitleFillColor(WHITE)
    ROOT.gStyle.SetStatColor(WHITE)

    ROOT.PyConfig.IgnoreCommandLineOptions = False
    ROOT.gROOT.SetBatch(True)


_rootlogon()



"""Contains the implementation of the functions used in launcher.py
"""
import json
import os
from pathlib import Path
from typing import List, Tuple

from URANIE import DataServer, Launcher, Sampler
import ROOT

from . import utils, input_data, execution as exe

URANIE_TCODE_LAUNCHER = "uranie-launcher-unitary"
URANIE_TCODE_JSON = ".uranie-launcher-commands.json"

URANIE_FAILED_VALUE = "1.234567890e+00"
"""This is the default value used by Uranie when calculation failed."""


def create_data_server(outputs: input_data.Outputs) -> DataServer.TDataServer:
    """Create the data server for Uranie

    Parameters
    ----------
    outputs : input_data.Outputs
        Object containing the name of the outputs and all the informations about them.

    Returns
    -------
    DataServer.TDataServer
        TDataServer object.
    """

    title = f"Quantities = {[output.quantity_of_interest for output in outputs.outputs]}"
    t_data_server = DataServer.TDataServer(outputs.name, title)
    t_data_server.changeIteratorName("execution_index")
    return t_data_server


def set_inputs(inputs: input_data.Inputs, t_data_server: DataServer.TDataServer):
    """Defines the parameters of the distribution law for each uncertains values,
    and associate the variable_name to the flag used in file_flag.

    Parameters
    ----------
    inputs : input_data.Inputs
        Object containing all the infos about the uncertain parameters.
    t_data_server : DataServer.TDataServer
        A TDataServer object

    Raises
    ------
    ValueError
        Invalid distribution type if the variable distribution isn't 'Uniform' or 'TruncatedNormal'.
    """
    for _input in inputs.inputs:
        variable_name = _input.variable_name
        if isinstance(_input.distribution, input_data.Inputs.DistributionUniform):
            t_data_server.addAttribute(DataServer.TUniformDistribution(
                variable_name, _input.distribution.lower_bound, _input.distribution.upper_bound))

        elif isinstance(_input.distribution, input_data.Inputs.DistributionTruncatedNormal):
            mean = _input.distribution.mean
            std_dev = _input.distribution.standard_deviation
            t_data_server.addAttribute(DataServer.TNormalDistribution(
                variable_name, mean, std_dev))
            t_data_server.getAttribute(variable_name).setBounds(
                _input.distribution.lower_bound, _input.distribution.upper_bound)

        else:
            raise ValueError("Invalid distribution type: "
                             f"{_input.distribution.__class__.__name__}")

        t_data_server.getAttribute(variable_name).setFileFlag(str(inputs.file_flag), _input.flag)


def generate_sample(propagation: input_data.Propagation,
                    outputs: input_data.Outputs,
                    t_data_server: DataServer.TDataServer,
                    output_directory: Path
                    ) -> Sampler.TQMC or Sampler.TSampling:
    """Definition of the experimental plan

    Parameters
    ----------
    propagation : input_data.Propagation
        Object containing the name of the propagation method used and the sample size.
    outputs : input_data.Outputs
        Object containing the name of the experimental design file name.
    t_data_server : DataServer.TDataServer
        A TDataServer object
    output_directory: Path
        Name of the directory where are stored the results
        of the uncertainty quantification calculation
    Returns
    -------
    Sampler.TQMC or Sampler.TSampling
        t_sampler object containing the experimental plan.

    Raises
    ------
    ValueError
        Invalid sampling method if the sampling method is not 'SRS' or 'Sobol'.
    """
    if propagation.sampling_method == propagation.SOBOL:
        # here I considered that canvas was deterministic
        t_sampler = Sampler.TQMC(
            t_data_server, propagation.sampling_method, propagation.sample_size)
    elif propagation.sampling_method == propagation.SRS:
        t_sampler = Sampler.TSampling(
            t_data_server, propagation.sampling_method, propagation.sample_size)
    else:
        raise ValueError(f"Unknown sampling method: {propagation.sampling_method}")

    t_sampler.generateSample()

    t_data_server.exportData(str(output_directory / outputs.experimental_design_filename))

    return t_sampler


def visualisation(t_data_server: DataServer.TDataServer):  # pragma: no cover
    """Plot a graph if the visualization variable is equal to True

    Parameters
    ----------
    t_data_server : DataServer.TDataServer
        A TDataServer object
    """
    ROOT.gROOT.SetBatch(False)  # pylint: disable=no-member
    canvas = ROOT.TCanvas()  # pylint: disable=no-member
    canvas.Clear()
    t_data_server.drawPairs()
    canvas.Draw()


def set_outputs(outputs: input_data.Outputs) -> List[Launcher.TOutputFileRow]:
    """Prepares the unitary aggregated outputs file.

    Parameters
    ----------
    outputs : input_data.Outputs
        Object containing the name of the outputs and all the informations about them.

    Returns
    -------
    List[Launcher.TOutputFileRow]
        List of file that will contain the unitary aggregated outputs
    """
    # The output files of the code
    t_output_files = []

    # The attribute in the output file
    for output in outputs.outputs:
        t_output_file = Launcher.TOutputFileRow(output.filename)
        for header in output.headers:
            t_output_file.addAttribute(DataServer.TAttribute(header))
        t_output_files.append(t_output_file)
    return t_output_files


def create_launcher(commands_to_execute: List[Tuple[str, List]],
                    t_data_server: DataServer.TDataServer,
                    output_directory: Path,
                    t_output_files: List[Launcher.TOutputFileRow]
                    ) -> Launcher.TLauncher:
    """Create the launcher for Uranie

    Parameters
    ----------
    commands_to_execute: List[Tuple[str, List]]
        Name of the script or the command which URANIE will execute with all its arguments.
    t_data_server : DataServer.TDataServer
        A TDataServer object
    output_directory : Path
        Name of the directory where are stored the results
        of the uncertainty quantification calculation
    t_output_files : List[Launcher.TOutputFileRow]
        List of files that will contain the unitary aggregated outputs

    Returns
    -------
    Launcher.TLauncher
        t_launcher object.
    """
    commands_json_file = output_directory / URANIE_TCODE_JSON
    with open(commands_json_file, 'w', encoding='utf-8') as tcode_file:
        json.dump(commands_to_execute, tcode_file, indent=4)

    command = [URANIE_TCODE_LAUNCHER, str(commands_json_file), "> log.out 2> log.err"]
    if utils.get_log_level() >= utils.DEBUG:
        command.insert(1, "--debug")
    t_code = Launcher.TCode(t_data_server, " ".join(command))

    # we add the output file of the code
    for t_output_file in t_output_files:
        t_code.addOutputFile(t_output_file)

    return Launcher.TLauncher(t_data_server, t_code)


def run_calculations(execution: exe.Execution,
                     t_launcher: Launcher.TLauncher,
                     output_directory: Path) -> Path:
    """Launch the calculation according to the appropriate mode (on your desktop or on a cluster).

    Parameters
    ----------
    execution: execution.Execution
        Object containing all the infos about how to execute the calculations.
    t_launcher : Launcher.TLauncher
        t_launcher object.
    output_directory : Path
        Name of the directory where are stored the results
        of the uncertainty quantification calculation

    Raises
    ------
    ValueError
        Invalid execution mode if it's different from 'desktop' or 'cluster'.
    ValueError
        execution.clean and output_directory == execution.working_directory is not possible.
    """

    # To back up all directories
    t_launcher.setSave(execution.save)
    t_launcher.setClean(execution.clean)

    # Uranie working directory
    t_launcher.setWorkingDirectory(str(execution.working_directory))

    if output_directory == execution.working_directory and execution.clean:
        raise ValueError(
            "execution.clean is True and output_directory == "
            "execution.working_directory is not possible.")

    if isinstance(execution, exe.ExecutionLocal):
        if execution.nb_jobs == 1:  # Launching code on a single processor
            # For display during runing
            if execution.visualization:  # pragma: no cover
                t_launcher.setVarDraw("max:initial_power", "", "")  # FIXME lie a integration_bench
            t_launcher.run()
        else:
            t_launcher.run(f"localhost={execution.nb_jobs}")

    # elif isinstance(execution, exe.ExecutionSlurm):
    #     pass
    #     #sbatch -n <nb de tâches> -c <nb de cœurs par tâche> -p <partition> --qos=<qos>
    #     # -A <account> -t <walltime HH:MM:SS> -J <jobname> -o <fichier de sortie>
    #     # -e <fichier d’erreur> --mail-user=<email> --mail-type=<BEGIN|END|FAIL|ALL>
    #     # -w <liste des nœuds> myscript.sh
    else:
        raise ValueError(f"Invalid execution mode: {execution.__class__.__name__}")


def _count_of_failed_calculations(uranie_work_dir: Path, unitary_result_filename: str):
    """ Counts the number of failed calculations.

    Parameters
    ----------
    uranie_work_dir : Path
        Absoulut path to the URANIE output directory.
    unitary_result_filename : str
        Name of the file containing the result of the
        aggregations functions for an unitary calculation

    Returns
    -------
    int
        Number of failed calculations.
    """
    list_directories = os.listdir(uranie_work_dir)
    list_directories.remove('UranieResults')
    nb_fail = 0
    for directory in list_directories:
        uranie_results_file = uranie_work_dir / directory / unitary_result_filename
        utils.debug(f"Check output {uranie_results_file}: {uranie_results_file.is_file()}")
        if not uranie_results_file.is_file():
            nb_fail += 1

    return nb_fail


def save_calculations(propagation: input_data.Propagation,
                      execution: exe.Execution,
                      outputs: input_data.Outputs,
                      t_data_server: DataServer.TDataServer,
                      output_directory: Path) -> Tuple[Path, int]:
    """Save calculation results in one file for valid calculations,
    and in another for failed calculations (if there is some).

    Parameters
    ----------
    propagation : input_data.Propagation
        Object containing the name of the propagation method used.
    execution: execution.Execution
        Object containing all the infos about how to execute the calculations.
    outputs : input_data.Outputs
        Object containing the outputs of the study to recover output filename.
    t_data_server : DataServer.TDataServer
        A TDataServer object.
    output_directory: Path
        Name of the directory where are stored the results
        of the uncertainty quantification calculation

    Returns
    -------
    Tuple[Path, int]
        Path to output file defined as ``output_directory/outputs.output_filename``
        and nb of failed.
    """

    output0 = outputs.outputs[0]
    nb_fail = _count_of_failed_calculations(execution.working_directory, output0.filename)

    ascii_filepath = output_directory / outputs.output_filename

    # Some succeded
    if nb_fail < propagation.sample_size:
        t_data_server.exportData(
            str(ascii_filepath), "*", f"{output0.headers[0]}!={URANIE_FAILED_VALUE}")

    if utils.get_log_level() >= utils.INFO:
        t_data_server.scan()

    # Some failed
    if nb_fail > 0:
        utils.info(f"\033[1;31m{nb_fail} over {propagation.sample_size} calculation(s) "
                   "failed !\033[0m")
        utils.debug(f"export failed simulations to {output_directory/outputs.failed_filename}.")
        utils.debug(f" -> condition: {output0.headers[0]}=={URANIE_FAILED_VALUE}")
        t_data_server.exportData(str(output_directory/outputs.failed_filename), "*",
                                 f"{output0.headers[0]}=={URANIE_FAILED_VALUE}")
    else:
        utils.info(f"\033[1;32mAll the {propagation.sample_size} calculation(s) "
                   "have succeeded !\033[0m")

    return ascii_filepath, nb_fail



""" Module used to set uncertainty data into an understandable format for uranie_launcher.
"""
from pathlib import Path
from typing import List


class Inputs():
    """ Class containing all the info about all the uncertain parameters.
    """
    class Distribution():  # pylint: disable=too-few-public-methods
        """ Class containing the info about the distribution for a specific uncertain parameter.
        """
        def __init__(self):
            pass

        # TODO string conversion
        # def __str__(self) -> str:
        #     return ''

    class DistributionUniform(Distribution):
        """Define the lower and upper bounds of a uniform distribution.

        Parameters
        ----------
        lower_bound : float
            lower bound of the distribution
        upper_bound : float
            upper bound of the distribution
        """
        def __init__(self, lower_bound, upper_bound):
            super().__init__()
            if lower_bound < upper_bound:
                self._lower_bound = lower_bound
                self._upper_bound = upper_bound
            else:
                raise ValueError("The upper bound have to be greater than the lower one !")

        @property
        def lower_bound(self) -> float:
            """Returns info about the distribution

            Returns
            -------
            float
                lower bound of the distribution
            """
            return self._lower_bound

        @property
        def upper_bound(self) -> float:
            """Returns info about the distribution

            Returns
            -------
            float
                upper bound of the distribution
            """
            return self._upper_bound

    class DistributionTruncatedNormal(Distribution):
        """ Set the arguments needed to define a truncated normal distribution.

        Parameters
        ----------
        lower_bound : float
            lower bound of the distribution
        upper_bound : float
            upper bound of the distribution
        mean : float
            mean of the distribution
        standard_deviation : float
            standard deviation of the distribution
        """
        def __init__(self, lower_bound, upper_bound, mean, standard_deviation):
            super().__init__()
            if lower_bound < upper_bound:
                self._lower_bound = lower_bound
                self._upper_bound = upper_bound
                self._mean = mean
                self._standard_deviation = standard_deviation
            else:
                raise ValueError("The upper bound have to be greater than the lower one !")

        @property
        def lower_bound(self) -> float:
            """Returns all info about the distribution

            Returns
            -------
            float
                lower bound of the distribution
            """
            return self._lower_bound

        @property
        def upper_bound(self) -> float:
            """Returns all info about the distribution

            Returns
            -------
            float
                upper bound of the distribution
            """
            return self._upper_bound

        @property
        def mean(self) -> float:
            """Returns all info about the distribution

            Returns
            -------
            float
                mean
            """
            return self._mean

        @property
        def standard_deviation(self) -> float:
            """Returns all info about the distribution

            Returns
            -------
            float
                standard deviation
            """
            return self._standard_deviation

    class Input():
        """ Class containing all the info about one specific parameter of the uncertain parameters.
        """
        def __init__(self, variable_name: str, distribution: 'Inputs.Distribution'):
            """ Set all the info about one specific parameter of the uncertain parameters.

            Parameters
            ----------
            variable_name : str
                Name of the uncertain variable
            distribution : Inputs.Distribution
                Is the distribution_info object previously defined
            """
            self._variable_name = variable_name
            self._distribution = distribution
            self._flag_delimiter = '@'

        @property
        def variable_name(self) -> str:
            """Returns the name of the uncertain value

            Returns
            -------
            str
                variable name
            """
            return self._variable_name

        @property
        def distribution(self) -> 'Inputs.Distribution':
            """Returns all info about the distribution

            Returns
            -------
            'Inputs.Distribution'
                distribution
            """
            return self._distribution

        @property
        def flag(self) -> str:
            """Returns the flag

            Returns
            -------
            str
                flag
            """
            return f"{self._flag_delimiter}{self._variable_name}{self._flag_delimiter}"

    def __init__(self):
        self._inputs = []
        self._file_flag = ""

    def add_input(self, _input: Input):
        """Adds an input class object into an inputs class list which contains all information
        about all uncertain parameters.

        Parameters
        ----------
        _input : Input
            Object containing all the info about one specific parameter of the uncertain parameters.
        """
        self._inputs.append(_input)

    @property
    def inputs(self) -> List[Input]:
        """Returns the _inputs list that contains all info about each uncertain parameter

        Returns
        -------
        list
            inputs
        """
        return self._inputs

    def set_file_flag(self, file_flag: Path):
        """ Set the name of the file containing the balise_flag which will be replaced
        by a specific value depending on the chosen distribution.

        Parameters
        ----------
        file_flag : Path
            Should be something like : <file_name>_Balise.<ext>
        """
        self._file_flag = file_flag

    @property
    def file_flag(self) -> str:
        """Returns the name of the file containing the balise

        Returns
        -------
        str
            file flag
        """
        return self._file_flag


class Propagation():
    """ Class defining an object containing the name
    of the propagation method used and the sample size.
    """

    SOBOL = "Sobol"
    """Sobol is blablabla... TODO"""

    SRS = "SRS"
    """SRS is blablabla... TODO"""

    def __init__(self, sampling_method: str, sample_size: int):
        if sample_size <= 0:
            raise ValueError("sample_size must be greater than 0")

        self._sampling_method = sampling_method
        self._sample_size = sample_size

    @property
    def sampling_method(self) -> str:
        """Returns the sampling method

        Returns
        -------
        str
            sampling method name
        """
        return self._sampling_method

    @property
    def sample_size(self) -> int:
        """Get the sample size variable

        Returns
        -------
        int
            sample size
        """
        return self._sample_size


class Outputs():
    """ Class defining an object containing all the informations about the outputs.
    """

    class Output():
        """ Class containing all the info about one specific output.
        """
        def __init__(self, headers: List[str], quantity_of_interest: str, filename: str):
            """Constructor.

            Parameters
            ----------
            headers : List[str]
                Names of the aggregation functions applied on the output
            quantity_of_interest : str
                Names of the output
            filename : str
                Name of the output file, if any
            """
            self._headers = headers
            self._quantity_of_interest = quantity_of_interest
            self._filename = filename

        @property
        def headers(self) -> List[str]:
            """Returns the dictionary containing all the chosen aggregation functions

            Returns
            -------
            list
                headers
            """
            return self._headers

        @property
        def quantity_of_interest(self) -> str:
            """Returns the name of the quantity of interest

            Returns
            -------
            str
                quantity_of_interest
            """
            return self._quantity_of_interest

        @property
        def filename(self) -> str:
            """Returns the name of the output file, if any

            Returns
            -------
            str
                file name
            """
            return self._filename

    def __init__(self, name):
        self._outputs = []
        self._name = name

    def add_output(self, output: 'Outputs.Output'):
        """ Set all the info about one specific output.

        Parameters
        ----------
        output : Outputs.Output
            output
        """
        self._outputs.append(output)

    @property
    def outputs(self) -> List[Output]:
        """Returns a list of the outputs containing all the info about each output

        Returns
        -------
        List[Output]
            outputs
        """
        return self._outputs

    @property
    def name(self) -> str:
        """Returns the name of the output

        Returns
        -------
        str
            tds_name
        """
        return self._name

    @property
    def experimental_design_filename(self) -> str:
        """Returns the name of the experimental design file

        Returns
        -------
        str
            ``self.name+'_exp_design.dat')``
        """
        return f"{self.name}_exp_design.dat"

    @property
    def output_filename(self) -> str:
        """Returns the name of the output file

        Returns
        -------
        str
            ``self.name+'_output.dat')``
        """
        return f"{self.name}_output.dat"

    @property
    def failed_filename(self) -> str:
        """Returns the name of the failed file

        Returns
        -------
        str
            ``self.name+'_failed.dat')``
        """
        return f"{self.name}_failed.dat"



"""Tests ``launcher`` module."""

import json
from pathlib import Path
import shutil

import pytest

from uranie_launcher import execution, input_data, launcher

from . import program_tester


## _run_unitary
def test_program_tester():
    """Test test program"""

    output_dirname = Path(__file__).absolute().parent / "test_program_tester"
    output_dirname.mkdir(parents=False, exist_ok=True)
    data_filepath = output_dirname / "input_for_tests.json"

    data_filepath.write_text(json.dumps({"uncertain_variable": 250}), encoding="utf-8")
    program_tester.main_unitary_calculation([str(data_filepath), str(output_dirname)])
    assert (output_dirname / "result").is_file()


    data_filepath.write_text(json.dumps({"uncertain_variable": 249}), encoding="utf-8")
    with pytest.raises(ZeroDivisionError):
        program_tester.main_unitary_calculation([str(data_filepath), str(output_dirname)])


def test_run_unitary():
    """Test unitary"""


## launcher
def test_launcher(commands_to_execute,
                  data_input_inputs_distribution_uniform,
                  data_input_outputs):
    """Test execute_uranie"""

    output_dirname = Path(__file__).absolute().parent / "test_launcher"
    if output_dirname.exists():
        shutil.rmtree(output_dirname)
    output_dirname.mkdir(parents=False, exist_ok=True)

    sample_size = 4
    propagation = input_data.Propagation(sampling_method=input_data.Propagation.SOBOL,
                                         sample_size=sample_size)

    exe = execution.ExecutionLocal(
        working_directory=output_dirname / "unitary", nb_jobs=sample_size)

    ascii_filepath, nb_failed = launcher.execute_uranie(
        commands_to_execute=commands_to_execute,
        inputs=data_input_inputs_distribution_uniform,
        propagation=propagation,
        outputs=data_input_outputs,
        execution=exe,
        output_directory=output_dirname)

    assert (
        nb_failed == 0 and
        ascii_filepath.exists()
    )



# -*- coding: utf-8 -*-

"""Installation script of the package."""

import pathlib
from setuptools import find_packages, setup


here = pathlib.Path(__file__).parent.resolve()


def get_version():
    """Extract the package's version number from the ``VERSION`` file."""
    return (here / "src" / "uranie_launcher" / "VERSION").read_text(encoding="utf-8").strip()


def get_long_description():
    """Extract README content"""
    return (here / "README.md").read_text(encoding="utf-8")


setup(
    name="uranie-launcher",
    version=get_version(),
    author="Clément STUTZ",
    author_email="clement.stutz@cea.fr",
    description="uranie-launcher help you to run your calculation scripts with Uranie",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={
        'console_scripts': [
            'uranie-launcher-unitary=uranie_launcher._run_unitary:run_unitary',
            'test-run-unitary-calculation=tests.program_tester:run_unitary_calculation',
        ]},
    install_requires=[
                    #   "URANIE",
                    #   "ROOT"
                      ],
    extras_require={
            'dev' : ["pytest"]
        },
    package_data={
        "uranie_launcher": [
            "VERSION",
        ],
    },
    project_urls={  # Optional
        "Bug Reports": "https://github.com/clementstutz/uranie-launcher/issues",
        "Source": "https://github.com/clementstutz/uranie-launcher",
    },
    )



"""Tests ``input_data`` module."""

from typing import List

import pytest

from uranie_launcher import input_data


## input_data
def test_distribution_uniform():
    """Test distribution"""

    lower_bound = 0
    upper_bound = 1
    distribution = input_data.Inputs.DistributionUniform(lower_bound, upper_bound)

    assert (
        isinstance(distribution, input_data.Inputs.DistributionUniform) and
        distribution.lower_bound == lower_bound and
        distribution.upper_bound == upper_bound
        )


def test_distribution_uniform_raise():
    """Test distribution"""

    lower_bound = 2
    upper_bound = 1

    with pytest.raises(ValueError) as error:
        input_data.Inputs.DistributionUniform(lower_bound, upper_bound)
    assert "The upper bound have to be greater than the lower one !" in str(error.value)


def test_distribution_truncated_normal():
    """Test distribution"""

    lower_bound = 0
    upper_bound = 1
    mean = 1
    standard_deviation = 0.05
    distribution = input_data.Inputs.DistributionTruncatedNormal(lower_bound,
                                                                 upper_bound,
                                                                 mean,
                                                                 standard_deviation)

    assert (
        isinstance(distribution, input_data.Inputs.DistributionTruncatedNormal) and
        distribution.lower_bound == lower_bound and
        distribution.upper_bound == upper_bound and
        distribution.mean == mean and
        distribution.standard_deviation == standard_deviation
        )


def test_distribution_truncated_normal_raise():
    """Test inputs"""

    lower_bound = 2
    upper_bound = 1
    mean = 1
    standard_deviation = 0.05

    with pytest.raises(ValueError) as error:
        input_data.Inputs.DistributionTruncatedNormal(lower_bound,
                                                      upper_bound,
                                                      mean,
                                                      standard_deviation)
    assert "The upper bound have to be greater than the lower one !" in str(error.value)


def test_one_input(simulation_variable):
    """Test inputs"""

    lower_bound = 0
    upper_bound = 1
    distribution = input_data.Inputs.DistributionUniform(lower_bound=lower_bound,
                                                         upper_bound=upper_bound)

    input_1 = input_data.Inputs.Input(simulation_variable, distribution)

    inputs = input_data.Inputs()
    inputs.add_input(input_1)

    assert (
        isinstance(inputs, input_data.Inputs) and
        isinstance(inputs.inputs, List) and
        len(inputs.inputs) == 1 and
        isinstance(inputs.inputs[0], input_data.Inputs.Input) and
        inputs.inputs[0] == input_1 and
        isinstance(inputs.inputs[0].distribution, input_data.Inputs.DistributionUniform) and
        inputs.inputs[0].variable_name == simulation_variable and
        inputs.inputs[0].distribution.lower_bound == lower_bound and
        inputs.inputs[0].distribution.upper_bound == upper_bound and
        inputs.inputs[0].flag == "@" + simulation_variable +"@"
        )


def test_two_input():
    """Test inputs"""

    variable_name_1 = "variable_1"
    lower_bound_1 = 0
    upper_bound_1 = 1
    distribution_1 = input_data.Inputs.DistributionUniform(lower_bound_1, upper_bound_1)

    variable_name_2 = "variable_2"
    lower_bound_2 = 0
    upper_bound_2 = 1
    mean = 1
    standard_deviation = 0.05
    distribution_2 = input_data.Inputs.DistributionTruncatedNormal(lower_bound_2,
                                                                   upper_bound_2,
                                                                   mean,
                                                                   standard_deviation)

    input_1 = input_data.Inputs.Input(variable_name_1, distribution_1)
    input_2 = input_data.Inputs.Input(variable_name_2, distribution_2)

    inputs = input_data.Inputs()
    inputs.add_input(input_1)
    inputs.add_input(input_2)

    assert (
        isinstance(inputs.inputs, List) and
        len(inputs.inputs) == 2 and
        isinstance(inputs.inputs[0], input_data.Inputs.Input) and
        inputs.inputs[0] == input_1 and
        isinstance(inputs.inputs[0].distribution, input_data.Inputs.DistributionUniform) and
        inputs.inputs[0].variable_name == variable_name_1 and
        inputs.inputs[0].distribution.lower_bound == lower_bound_1 and
        inputs.inputs[0].distribution.upper_bound == upper_bound_1 and
        inputs.inputs[0].flag == "@" + variable_name_1 +"@" and
        isinstance(inputs.inputs[1], input_data.Inputs.Input) and
        inputs.inputs[1] == input_2 and
        isinstance(inputs.inputs[1].distribution, input_data.Inputs.DistributionTruncatedNormal) and
        inputs.inputs[1].variable_name == variable_name_2 and
        inputs.inputs[1].distribution.lower_bound == lower_bound_2 and
        inputs.inputs[1].distribution.upper_bound == upper_bound_2 and
        inputs.inputs[1].distribution.mean == mean and
        inputs.inputs[1].distribution.standard_deviation == standard_deviation and
        inputs.inputs[1].flag == "@" + variable_name_2 +"@"
        )


def test_file_flag():
    """Test file flag"""

    inputs = input_data.Inputs()
    scenario_flag_filename = "some/path/to/uranie_tag_input.xml"
    inputs.set_file_flag(scenario_flag_filename)

    assert inputs.file_flag == scenario_flag_filename


def test_propagation():
    """Test propagation"""

    sampling_method = "name_of_sampling_method"
    sample_size = 4
    propagation = input_data.Propagation(sampling_method, sample_size)

    assert (
        isinstance(propagation, input_data.Propagation) and
        propagation.sampling_method == sampling_method and
        propagation.sample_size == sample_size
        )


def test_sample_size_zero_raises_value_error():
    """Test sample"""

    with pytest.raises(ValueError) as error:
        input_data.Propagation(sampling_method="name_of_sampling_method", sample_size=0)
    assert (
        "sample_size must be greater than 0" in str(error.value)
    )


def test_one_output(headers, quantity_of_interest):
    """Test outputs"""

    filename = "filename"
    output = input_data.Outputs.Output(filename=filename,
                                       headers=headers,
                                       quantity_of_interest=quantity_of_interest)

    assert (
        isinstance(output, input_data.Outputs.Output) and
        output.filename == filename and
        output.headers == headers and
        output.quantity_of_interest == quantity_of_interest
        )


def test_outputs(headers, quantity_of_interest, data_server_name):
    """Test outputs"""

    outputs = input_data.Outputs(data_server_name)
    outputs.add_output(input_data.Outputs.Output(headers, quantity_of_interest, "output_file"))

    assert (
        isinstance(outputs, input_data.Outputs) and
        outputs.name == data_server_name and
        isinstance(outputs.outputs, List) and
        isinstance(outputs.outputs[0], input_data.Outputs.Output) and
        outputs.outputs[0].headers == headers and
        outputs.outputs[0].quantity_of_interest == quantity_of_interest
        )


def test_two_output():
    """Test outputs"""

    headers = ['toto', 'tata']

    headers_1 = headers
    quantity_of_interest_1 = "Temperature"
    output_1 = input_data.Outputs.Output(headers_1, quantity_of_interest_1, "Temperature_output")

    headers_2 = headers[:-1]
    quantity_of_interest_2 = "Density"
    output_2 = input_data.Outputs.Output(headers_2, quantity_of_interest_2, "Density_output")

    name = "Name_of_the_study"
    outputs = input_data.Outputs(name)
    outputs.add_output(output_1)
    outputs.add_output(output_2)

    assert isinstance(outputs, input_data.Outputs)
    assert outputs.name == name
    assert isinstance(outputs.outputs, List)
    assert isinstance(outputs.outputs[0], input_data.Outputs.Output)
    assert outputs.outputs[0].headers == headers_1
    assert outputs.outputs[0].quantity_of_interest == quantity_of_interest_1
    assert isinstance(outputs.outputs[1], input_data.Outputs.Output)
    assert outputs.outputs[1].headers == headers_2
    assert outputs.outputs[1].quantity_of_interest == quantity_of_interest_2

    assert outputs.experimental_design_filename == f"{name}_exp_design.dat"
    assert outputs.output_filename == f"{name}_output.dat"
    assert outputs.failed_filename == f"{name}_failed.dat"



"""Module to load, parse and write data based on uranie file format (ascii or json).
"""
import csv
from datetime import datetime
import json
from pathlib import Path
from typing import Any, List


class Data():
    """Class containing the data."""

    class Types:
        """Class defining the type available in ``Data``."""

        STRING = "S"
        """Value of type ``str``."""
        DOUBLE = "D"
        """Value of type ``float``."""
        VECTOR = "V"
        """Value of type ``list``."""

        @staticmethod
        def convert(value: str, value_type: str) -> Any:
            """Convert from string to Uranie type

            Parameters
            ----------
            value : str
                Value as string.
            value_type : str
                Type of the value

            Returns
            -------
            Any
                Value casted as expected type
            """
            return {
                Data.Types.STRING: str,
                Data.Types.DOUBLE: float,
                Data.Types.VECTOR: lambda val: [float(v) for v in val.strip('][').split(',')],
            }[value_type](value)

        @staticmethod
        def check_type(value: Any, value_type: str, var_name: str, row_index: int):
            """Check the type of a value.

            Parameters
            ----------
            value : Any
                Value to test
            value_type : str
                Uranie ``Data.Types``
            var_name : str
                Name associated to the variable
            row_index : int
                Row in the Data

            Raises
            ------
            ValueError
                The type is not in ``Data.Types``.
            ValueError
                The type of an element of ``Data.Types.VECTOR`` is not float.
            """

            types = {
                Data.Types.STRING: str,
                Data.Types.DOUBLE: (float, int),
                Data.Types.VECTOR: list,
            }
            if not isinstance(value, types[value_type]):
                raise ValueError(
                    f"Type is not correct: found {type(value)},"
                    f"expected {types[value_type]} for '{var_name}' at row {row_index}.")

            if value_type == Data.Types.VECTOR:
                for elt, val in enumerate(value):
                    if not isinstance(val, float):
                        raise ValueError(
                            f"Element of vector is not correct: found {type(val)},"
                            f"expected float for '{var_name}' at row {row_index}, element {elt}.")

    class Header:
        """Class defining data header"""

        def __init__(self, name: str, value_type: str, value_unit: str) -> None:
            """Constructor.

            Parameters
            ----------
            name : str
                name of the variable
            value_type : str
                type among ``Data.Types``
            value_unit : str
                unit of the variable
            """
            self._name: str = name
            self._value_type: str = value_type
            self._value_unit: str = value_unit

        @property
        def name(self) -> str:
            """name of the variable"""
            return self._name

        @property
        def value_type(self) -> str:
            """type among ``Data.Types``"""
            return self._value_type

        @property
        def value_unit(self) -> str:
            """unit of the variable"""
            return self._value_unit

    def __init__(self, name: str, description: str, headers: List['Data.Header']):
        """Constructor.

        Parameters
        ----------
        name : str
            Name of the data
        description : str
            Description of the data
        headers : List[Data.Header]
            List of ``Data.Header`` to define the content
        """
        self._name: str = name
        self._description: str = description
        self._headers: List['Data.Header'] = headers
        self._values: List[List[float or str or List[float]]] = [
            [] for i in range(len(self._headers))]

    @property
    def name(self) -> str:
        """Name of the data"""
        return self._name

    @property
    def description(self) -> str:
        """Description of the data"""
        return self._description

    @property
    def headers(self) -> List['Data.Header']:
        """List of ``Data.Header`` defining the content"""
        return self._headers

    @property
    def names(self) -> List[str]:
        """List of header names"""
        return [header.name for header in self.headers]

    @property
    def types(self) -> List[str]:
        """List of header types"""
        return [header.value_type for header in self.headers]

    @property
    def units(self) -> List[str]:
        """List of header units"""
        return [header.value_unit for header in self.headers]

    @property
    def values(self) -> List[List[float or str or List[float]]]:
        """values per header"""
        return self._values

    @property
    def nb_rows(self):
        """Number of values per header"""
        return len(self.values[0])

    @property
    def nb_columns(self):
        """Number of header"""
        return len(self.headers)

    def add_values(self, values: List[float or str or List[float]]):
        """Add a row in the data.

        Parameters
        ----------
        values : List[float or str or List[float]]
            Value for each header

        Raises
        ------
        ValueError
            If type is not coherent with header
        """

        for index, value in enumerate(values):
            Data.Types.check_type(value=value,
                                  value_type=self.types[index],
                                  row_index=len(self._values[index]),
                                  var_name=self.names[index])
        for index, value in enumerate(values):
            self._values[index].append(value)

    def get_values(self, index) -> List[float or str or List[float]]:
        """Get a row in the data.

        Parameters
        ----------
        index : _type_
            Row index

        Returns
        -------
        List[float or str or List[float]]
            Value for each header
        """
        return [values[index] for values in self.values]


def data_to_csv(data: Data, filepath: Path):
    """Convert ``Data`` to csv.

    Parameters
    ----------
    data : Data
        data to dump
    filepath : Path
        Path to file
    """
    with filepath.open(mode='w', encoding='utf-8', newline='') as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(["NAME", data.name])
        csv_writer.writerow(["TITLE", data.description])
        csv_writer.writerow(["DATE", _get_date()])
        csv_writer.writerow(["COLUMN_NAMES"] + data.names)
        csv_writer.writerow(["COLUMN_TYPES"] + data.types)
        csv_writer.writerow(["COLUMN_TITLES"] + data.names)
        csv_writer.writerow(["COLUMN_UNITS"] + data.units)
        csv_writer.writerows([[index] + [str(values[index]) for values in data.values]
                              for index in range(data.nb_rows)])


def csv_to_data(filepath: Path) -> Data:
    """Convert csv to ``Data``.

    Parameters
    ----------
    filepath : Path
        Path to file

    Returns
    -------
    Data
        Loaded data
    """
    with filepath.open(mode='r', encoding='utf-8') as csv_file:

        matrix = []
        csv_reader = csv.reader(csv_file)
        for line_no, line in enumerate(csv_reader):
            if line_no == 0:
                name = line[1]
            elif line_no == 1:
                description = line[1]
            elif line_no == 2:
                pass  # date
            elif line_no == 3:
                names = line[1:]
            elif line_no == 4:
                types = line[1:]
            elif line_no == 5:
                pass  # titles
            elif line_no == 6:
                units = line[1:]
            else:
                matrix.append([Data.Types.convert(value, types[index])
                               for index, value in enumerate(line[1:])])

        data = Data(name=name,
                    description=description,
                    headers=[Data.Header(name=name,
                                         value_type=types[index],
                                         value_unit=units[index])
                             for index, name in enumerate(names)])
        for values in matrix:
            data.add_values(values)
        return data


def _get_date():
    return datetime.now().strftime("%a %b %d %H:%M:%S %Y")  # "Fri Oct 28 10:41:44 2016"


def data_to_json(data: Data, filepath: Path):
    """Convert ``Data`` to json.

    Parameters
    ----------
    data : Data
        data to dump
    filepath : Path
        Path to file
    """
    dico = {
        "_metadata":
        {
            "_comment": "",
            "date": _get_date(),
            "short_names": data.names,
            "table_description": data.description,
            "table_name": data.name,
            "types": data.types,
            "units": data.units
        },
        "items":
        [{name: values[index]
            for name, values in zip(data.names, data.values)}
            for index in range(data.nb_rows)]
    }
    filepath.write_text(json.dumps(dico, indent=2), encoding='utf-8')


def json_to_data(filepath: Path) -> Data:
    """Convert json to ``Data``.

    Parameters
    ----------
    filepath : Path
        Path to file

    Returns
    -------
    Data
        Loaded data
    """
    dico = json.loads(filepath.read_text(encoding='utf-8'))
    data = Data(name=dico["_metadata"]["table_name"],
                description=dico["_metadata"]["table_description"],
                headers=[Data.Header(name=name,
                                     value_type=dico["_metadata"]["types"][index],
                                     value_unit=dico["_metadata"]["units"][index])
                         for index, name in enumerate(dico["_metadata"]["short_names"])])
    for item in dico["items"]:
        data.add_values([item[header.name] for header in data.headers])
    return data


def data_to_ascii(data: Data, filepath: Path):
    """Convert ``Data`` to 'Salome Table' as defined by Uranie.

    Parameters
    ----------
    data : Data
        data to dump
    filepath : Path
        Path to file
    """
    filepath.write_text(f"""
#NAME: {data.name}
#TITLE: {data.description}
#DATE: {_get_date()}
#COLUMN_NAMES: {'|'.join(data.names)}
#COLUMN_TYPES: {'|'.join(data.types)}
#COLUMN_TITLES: {'|'.join(data.names)}
#COLUMN_UNITS: {'|'.join(data.units)}

""" + "\n".join([" ".join([str(values[index]).replace(' ', '') for values in data.values])
                 for index in range(data.nb_rows)]) + "\n", encoding='utf-8')


def ascii_to_data(filepath: Path) -> Data:  # pylint: disable=too-many-branches  # noqa: C901
    """Convert 'Salome Table' as defined by Uranie to ``Data``.

    Parameters
    ----------
    filepath : Path
        Path to file

    Returns
    -------
    Data
        Loaded data
    """

    name = ""
    description = ""
    names = []
    types = []
    units = []
    matrix = []
    for line in filepath.read_text(encoding='utf-8').splitlines():
        if not line.strip():
            continue
        if line.startswith("#NAME:"):
            name = line.replace('#NAME:', '').strip()
        elif line.startswith("#TITLE:"):
            description = line.replace('#TITLE:', '').strip()
        elif line.startswith("#DATE:"):
            pass
        elif line.startswith("#COLUMN_NAMES:"):
            names = [n.strip() for n in line.replace('#COLUMN_NAMES:', '').strip().split('|')]
        elif line.startswith("#COLUMN_TYPES:"):
            types = [n.strip() for n in line.replace('#COLUMN_TYPES:', '').strip().split('|')]
        elif line.startswith("#COLUMN_TITLES:"):
            pass
        elif line.startswith("#COLUMN_UNITS:"):
            units = [n.strip() for n in line.replace('#COLUMN_UNITS:', '').strip().split('|')]
        else:
            matrix.append([Data.Types.convert(value, types[index])
                           for index, value in enumerate(line.split())] if types else
                          [float(value) for value in line.split()])

    if not names:
        names = [str(index) for index, _ in enumerate(matrix[0])]
    if not types:
        types = ["D"] * len(names)
    if not units:
        units = [""] * len(names)

    data = Data(name=name,
                description=description,
                headers=[Data.Header(name=name,
                                     value_type=types[index],
                                     value_unit=units[index])
                         for index, name in enumerate(names)])
    for values in matrix:
        data.add_values(values)
    return data



""" Script for launching a study with URANIE.
    It reads study parameters from input_data subclasses,
    and runs the script (or the command) commands_to_execute,
    which executes the individual calculation as many times as required.
"""

from pathlib import Path
from typing import List, Tuple
from uranie_launcher import _data_2_uranie as d2u
from uranie_launcher import input_data, execution as exe


def execute_uranie(commands_to_execute: List[Tuple[str, List]],
                   inputs: input_data.Inputs,
                   propagation: input_data.Propagation,
                   outputs: input_data.Outputs,
                   execution: exe.Execution,
                   output_directory: Path) -> Tuple[Path, int]:
    """ Use URANIE to genetrate the experiment plan based on inputs, then launch calculations.

    Parameters
    ----------
    commands_to_execute: List[Tuple[str, List]]
        Name of the script or the command which URANIE will execute with all its arguments.
    inputs: input_data.Inputs
        Object containing all the infos about the uncertain parameters.
    propagation: input_data.Propagation
        Object containing the name of the propagation method used.
    outputs: input_data.Outputs
        Object containing the name of the outputs and all the informations about them.
    execution: execution.Execution
        Object containing all the infos about how to execute the calculations.
    output_directory: Path
        Name of the directory where are stored the results
        of the uncertainty quantification calculation

    Returns
    -------
    Path
        Path to output file defined as ``output_directory/outputs.output_filename``.
    """

    t_data_server = d2u.create_data_server(outputs)

    d2u.set_inputs(inputs, t_data_server)

    d2u.generate_sample(propagation, outputs, t_data_server, output_directory)

    if execution.visualization:  # pragma: no cover
        d2u.visualisation(t_data_server)

    t_output_files = d2u.set_outputs(outputs)

    # Code instantiation
    t_launcher = d2u.create_launcher(commands_to_execute,
                                     t_data_server,
                                     output_directory,
                                     t_output_files)

    d2u.run_calculations(execution, t_launcher, output_directory)

    ascii_filepath, nb_failed = d2u.save_calculations(propagation,
                                                      execution,
                                                      outputs,
                                                      t_data_server,
                                                      output_directory)

    if execution.visualization:  # pragma: no cover
        d2u.visualisation(t_data_server)
        input('Type Enter to quit the program...')

    return ascii_filepath, nb_failed
